// **************************************************
// *  CHOCO: an open-source Constraint Programming  *
// *     System for Research and Education          *
// *                                                *
// *    contributors listed in choco.Entity.java    *
// *           Copyright (C) F. Laburthe, 1999-2006 *
// **************************************************
package choco;

import choco.bool.BoolConstraint;
import choco.bool.CompositeConstraint;
import choco.mem.IStateBool;
import choco.prop.*;

/**
 * An abstract class for all implementations of listeners
 */
public abstract class AbstractConstraint extends AbstractEntity implements Propagator {

  /**
   * The priority of the constraint.
   */

  protected int priority;


  /**
   * The constraint <i>awake</i> var attached to the constraint.
   */

  protected ConstraintEvent constAwakeEvent;


  /**
   * A field for attaching additional data util (useful for PaLM)
   */
  protected ConstraintPlugin hook;


  /**
   * a field for storing whether the constraint is active or not
   */
  protected IStateBool active;

  /**
   * Constraucts a constraint with the priority 0.
   */

  public AbstractConstraint() {
    this(0);
  }


  /**
   * Constructs a constraint with the specified priority.
   *
   * @param priority The wished priority.
   */

  public AbstractConstraint(int priority) {
    this.priority = priority;
    this.constAwakeEvent = new ConstraintEvent(this, false, priority);
  }


  /**
   * Returns the constraint plugin. Useful for extending the solver.
   */

  public ConstraintPlugin getPlugIn() {
    return hook;
  }

  /**
   * Set the ConstraintPlugin of the constraint. The constraint plugin
   * gives access to the dependency net of each constraint, it can be useful
   * to design search heuristic based on explanation.
   *
   * @param hook
   */
  public void setPlugIn(ConstraintPlugin hook) {
    this.hook = hook;
  }

  /**
   * Returns the constraint awake var attached to the constraint.
   */

  public PropagationEvent getEvent() {
    return constAwakeEvent;
  }


  /**
   * Initial propagation of the constraint.
   */

  public void constAwake(boolean isInitialPropagation) {
    ((ChocEngine) getProblem().getPropagationEngine())
        .postConstAwake(this, isInitialPropagation);
  }


  /**
   * Returns the priority.
   */

  public int getPriority() {
    return priority;
  }


  /**
   * Default propagation on variable revision: full constraint re-propagation.
   */

  public void awakeOnVar(int idx) throws ContradictionException {
    propagate();
  }


  /**
   * Default initial propagation: full constraint re-propagation.
   */

  public void awake() throws ContradictionException {
    propagate();
  }


  /**
   * Retrieve the problem of the constraint.
   */

  // ??????????? add the case when the constraint c involves only
  // constant variables, considered as belonging to CURRENT_PB
  public AbstractProblem getProblem() {
    int nVars = getNbVars();
    for (int i = 0; i < nVars; i++) {
      Var v = null;
      v = getVar(i);
      if (!(v.getProblem() == null))
        return v.getProblem();
    }
    return null;
  }


  /**
   * Returns the constraint that is in the network and involving the
   * current constraint as a subconstraint.
   */

  Constraint getRootConstraint() {
    Constraint rootConstraint = null;
    rootConstraint = (Constraint) getVar(0).getConstraint(getConstraintIdx(0));
    return rootConstraint;
  }


  /**
   * Un-freezing a constraint (this is useful for mimicking dynamic
   * constraint posts...).
   */

  public void setActive() {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      if (!(isActive())) {
        active.set(true);
        if (hook != null) hook.activateListener();
        constAwake(true);
      }
    }
  }


  /**
   * Freezing a constraint (this is useful for backtracking when mimicking
   * dynamic constraint posts...).
   */

  public void setPassive() {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      if (isActive()) {
        active.set(false);
        ConstraintEvent evt = constAwakeEvent;
        EventQueue q = ((ChocEngine) getProblem().getPropagationEngine()).getQueue(evt);
        if (this.hook != null) this.hook.deactivateListener();
        q.remove(evt);
      }
    }
  }

  /**
   * records that a constraint is now entailed (therefore it is now useless to propagate it again)
   */
  public void setEntailed() {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      setPassive();
    } else {
      CompositeConstraint root = (CompositeConstraint) rootConstraint;
      int varOffset = root.getGlobalVarIndex(this, 0);
      ((BoolConstraint) root).setSubConstraintStatus(this, true, varOffset);
    }
  }

  /**
   * Removes a constraint from the network.
   * Beware, this is a permanent removal, it may not be backtracked
   */

  public void delete() {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      setPassive();
      getProblem().eraseConstraint(this);
      for (int i = 0; i < getNbVars(); i++) {
        AbstractVar v = (AbstractVar) getVar(i);
        v.eraseConstraint(this);
      }
    }
  }

  /**
   * raise a contradiction during propagation when the constraint can definitely not be satisfied given the current domains
   */
  public void fail() throws ContradictionException {
    getProblem().getPropagationEngine().raiseContradiction(this);
  }

  public Boolean isEntailed() {
    if (isCompletelyInstantiated()) {
      return Boolean.valueOf(isSatisfied());
    } else {
      return null;
    }

  }

  /**
   * Checks if the constraint is active (e.g. plays a role in the propagation phase).
   *
   * @return true if the constraint is indeed currently active
   */

  public boolean isActive() {
    return active.get();
  }

  /**
   * This function connects a constraint with its variables in several ways.
   * Note that it may only be called once the constraint
   * has been fully created and is being posted to a problem.
   * Note that it should be called only once per constraint.
   * This can be a dynamic addition (undone upon backtracking) or not
   */

  public void addListener(boolean dynamicAddition) {
    int n = getNbVars();
    for (int i = 0; i < n; i++) {
      setConstraintIndex(i, getVar(i).addConstraint(this, i, dynamicAddition));
    }
    if (this.hook != null) this.hook.addListener();
    active = this.getProblem().getEnvironment().makeBool(true);
  }


  // defaut implementation: returns the same numbering in a constraint and its counterpart.
  public int getVarIdxInOpposite(int i) {
    return i;
  }

  public AbstractConstraint opposite() {
    throw new UnsupportedOperationException();
  }

  public int connectVar(AbstractVar v, int j, boolean dynamicAddition) {
    int cidx = v.addConstraint(this, j, dynamicAddition);
    setConstraintIndex(j, cidx);
    return cidx;
  }

  public boolean isEquivalentTo(Constraint compareTo) {
    return this.equals(compareTo);
  }

  public Object clone() throws CloneNotSupportedException {
    return super.clone();
  }

  /**
   * substitues all occurrences of a variable in a constraint by another variable
   *
   * @param oldvar the variable to be removed
   * @param newvar the variable to be introduced in place of the other
   * @return the number of occurrences that have been substituted
   */
  public int substituteVar(Var oldvar, Var newvar) {
    int nbSub = 0;
    int nbVars = this.getNbVars();
    for (int i = 0; i < nbVars; i++) {
      Var v = this.getVar(i);
      if (this.getVar(i) == oldvar) {
        this.setVar(i, newvar);
        nbSub++;
      }
    }
    return nbSub;
  }
}
