// **************************************************
// *  CHOCO: an open-source Constraint Programming  *
// *     System for Research and Education          *
// *                                                *
// *    contributors listed in choco.Entity.java    *
// *           Copyright (C) F. Laburthe, 1999-2006 *
// **************************************************
package choco;

import choco.integer.IntConstraint;
import choco.integer.IntDomainVar;
import choco.integer.IntVar;
import choco.mem.Environment;
import choco.prop.*;
import choco.real.RealVar;
import choco.util.IntIterator;
import choco.util.LightFormatter;

import java.util.Iterator;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.logging.StreamHandler;
import java.util.logging.Handler;
import java.security.AccessControlException;

public abstract class AbstractProblem extends AbstractModel {
  /**
   * Precision of the search for a real problem.
   */
  protected double precision = 1.0e-6;
  /**
   * Minimal width reduction between two propagations.
   */
  protected double reduction = 0.99;
  /**
   * Allows to know if the problem is feasible (null if it was not solved).
   */
  public Boolean feasible = null;
  /**
   * True if the problem was solved.
   */
  protected boolean solved = false;
  /**
   * The environment managing the backtrackable data.
   */
  protected Environment environment;
  /**
   * The propagation engine to propagate during solving.
   */
  protected AbstractPropagationEngine propagationEngine;
  /**
   * The object controlling the global search exploration
   */
  protected Solver solver;

  public AbstractProblem(Environment env) {
    super();
    this.solver = new Solver(this);
    this.propagationEngine = new ChocEngine(this);
    this.environment = env;
    this.constraints = env.makePartiallyStoredVector();
  }

  public AbstractProblem() {
    this(new Environment());
  }

  /**
   * Returns the memory environment used by the problem.
   */

  public final Environment getEnvironment() {
    return environment;
  }

  public int getIntVarIndex(IntDomainVar c) {
    return intVars.indexOf(c);
  }

  /**
   * retrieving the total number of constraints over integers
   *
   * @return the total number of constraints over integers in the problem
   */
  public final int getNbIntConstraints() {
    return this.constraints.size();
  }

  /**
   * <i>Network management:</i>
   * Retrieve a variable by its index (all search variables of
   * the problem are numbered in sequence from 0 on)
   *
   * @param i index of the variable in the problem
   * @deprecated
   */

  public final IntConstraint getIntConstraint(int i) {
    return (IntConstraint) constraints.get(i);
  }

  public Iterator getIntConstraintIterator() {
    return new Iterator() {
      IntIterator it = constraints.getIndexIterator();

      public boolean hasNext() {
        return it.hasNext();
      }

      public Object next() {
        return constraints.get(it.next());
      }

      public void remove() {
        throw new UnsupportedOperationException();
      }
    };
  }

  /**
   * As all Entity, must be able to return a Problem, returns itself.
   */

  public final AbstractProblem getProblem() {
    return this;
  }

  /**
   * test if the problem has been found to be feasible (there exist solutions) or not.
   *
   * @return Boolean.TRUE if a solution was found,
   *         Boolean.FALSE if the problem was proven infeasible,
   *         null otherwise
   */
  public final Boolean isFeasible() {
    return feasible;
  }

  public final boolean isConsistent() {
    int nbc = getNbIntConstraints();
    boolean consistent = true;
    for (int i = 0; (i < nbc && consistent); i++) {
      IntConstraint c = getIntConstraint(i);
      consistent = (consistent && c.isConsistent());
    }
    return consistent;
  }

  public String pretty() {
    StringBuffer buf = new StringBuffer("Pb[" + (getNbIntVars() + getNbRealVars() + getNbSetVars()) + " vars, " + getNbIntConstraints() + " cons]\n");
    buf.append(this.varsToString());
    buf.append(this.constraintsToString());
    return new String(buf);
  }

  public String varsToString() {
    StringBuffer buf = new StringBuffer("Pb[" + (getNbIntVars() + getNbRealVars() + getNbSetVars()) + " vars, " + getNbIntConstraints() + " cons]\n");
    buf.append("==== VARIABLES ====\n");
    for (int i = 0; i < getNbIntVars(); i++) {
      buf.append(getIntVar(i).pretty());
      buf.append("\n");
    }
    for (int i = 0; i < floatVars.size(); i++) {
      RealVar realVar = (RealVar) floatVars.get(i);
      buf.append(realVar.pretty());
    }
    for (int i = 0; i < setVars.size(); i++) {
      buf.append(getSetVar(i).pretty());
      buf.append("\n");
    }
    return new String(buf);
  }

  public String constraintsToString() {
    StringBuffer buf = new StringBuffer("Pb[" + (getNbIntVars() + getNbRealVars() + getNbSetVars()) + " vars, " + getNbIntConstraints() + " cons]\n");
    buf.append("==== CONSTRAINTS ====\n");
    IntIterator it = constraints.getIndexIterator();
    while (it.hasNext()) {
      int i = it.next();
      AbstractConstraint c = (AbstractConstraint) constraints.get(i);
      buf.append(c.pretty());
      buf.append("\n");
    }
    return new String(buf);
  }

  public String solutionToString() {
    StringBuffer buf = new StringBuffer();
    for (int i = 0; i < getNbIntVars(); i++) {
      IntVar v = getIntVar(i);
      if (v.isInstantiated()) {
        buf.append(v.toString());
        buf.append(", ");
      }
    }
    return new String(buf);
  }

  /**
   * removes (permanently) a constraint from the overall list of constraints of the problem
   */
  public void eraseConstraint(Constraint c) {
    constraints.remove(c);
    this.nbConstraint--;
  }

  public double getPrecision() {
    return precision;
  }

  public void setPrecision(double precision) {
    this.precision = precision;
  }

  public double getReduction() {
    return reduction;
  }

  public void setReduction(double reduction) {
    this.reduction = reduction;
  }

  /**
   * Returns the propagation engine associated to the problem
   */

  public PropagationEngine getPropagationEngine() {
    return propagationEngine;
  }

  /**
   * accessor returning the handler of the object responsible for the resolution of the problem.
   * It may be either a local search solver or a global search solver
   *
   * @return the object responsible for solving the problem
   */
  public final Solver getSolver() {
    return solver;
  }

  public abstract void post(Constraint c);

  public abstract void postCut(Constraint c);

  /**
   * <i>Propagation:</i>
   * Computes consistency on the problem (the problem may no longer
   * be consistent since the last propagation because of listeners
   * that have been posted and variables that have been reduced
   *
   * @throws ContradictionException
   */

  public void propagate() throws ContradictionException {
    PropagationEngine pe = getPropagationEngine();
    boolean someEvents = true;
    while (someEvents) {
      EventQueue q = pe.getNextActiveEventQueue();
      if (q != null) {
        q.propagateSomeEvents();
      } else
        someEvents = false;
    }
    assert(pe.checkCleanState());
  }

  /**
   * pushing one world on the stack
   */
  public void worldPush() {
    environment.worldPush();
  }

  /**
   * pushing one world on the stack
   */
  public void worldPop() {
    environment.worldPop();
    propagationEngine.flushEvents();
  }

  /**
   * Backtracks to a given level in the search tree.
   */
  public final void worldPopUntil(int n) {
    while (environment.currentWorld > n) {
      worldPop();
    }
  }

  /**
   * returning the index of the current worl
   */
  public final int getWorldIndex() {
    return environment.getWorldIndex();
  }

  public Boolean solve() {
    solver.firstSolution = true;
    solver.generateSearchSolver(this);
//    solver.getSearchSolver().incrementalRun();
    solver.launch();
    return isFeasible();
  }

  public Boolean nextSolution() {
    return solver.getSearchSolver().nextSolution();
  }

  /**
   * Displays all the runtime statistics.
   */
  public void printRuntimeSatistics() {
    (this.getSolver().getSearchSolver()).printRuntimeStatistics();
  }

  public abstract Boolean solve(boolean all);

  public Boolean solveAll() {
    solver.firstSolution = false;
    solver.generateSearchSolver(this);
//    solver.getSearchSolver().incrementalRun();
    solver.launch();
    return Boolean.TRUE;
  }

  /**
   * <i>Resolution:</i>
   * Searches for the solution minimizing the objective criterion.
   *
   * @param obj     The variable modelling the optimization criterion
   * @param restart If true, then a new search is restarted from scratch
   *                after each solution is found;
   *                otherwise a single branch-and-bound search is performed
   */

  public Boolean minimize(Var obj, boolean restart) {
    return optimize(false, obj, restart);
  }

  /**
   * <i>resolution:</i>
   * Searches for the solution maximizing the objective criterion.
   *
   * @param obj     The variable modelling the optimization criterion
   * @param restart If true, then a new search is restarted from scratch
   *                after each solution is found;
   *                otherwise a single branch-and-bound search is performed
   */
  public Boolean maximize(Var obj, boolean restart) {
    return optimize(true, obj, restart);
  }

  protected Boolean optimize(boolean maximize, Var obj, boolean restart) {
    solver.setDoMaximize(maximize);
    solver.setObjective(obj);
    solver.setRestart(restart);
    solver.setFirstSolution(false);
    solver.generateSearchSolver(this);
    solver.launch();
    return this.isFeasible();
  }


}
